package edu.ntnu.stud.chaosgame.controller.game;
import edu.ntnu.stud.chaosgame.controller.game.ChaosGame;
import edu.ntnu.stud.chaosgame.model.data.Vector2D;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.*;

// Step 1: Create a Callable tasks
class ChaosGameTask implements Callable<Void> {
  private final int start;
  private final int end;
  private final ChaosGame game;

  public ChaosGameTask(ChaosGame game, int start, int end) {
    this.game = game;
    this.start = start;
    this.end = end;
  }

  @Override
  public Void call() {
    for (int i = start; i < end; i++) {
      int j = game.getRandom().nextInt(game.getDescription().getTransforms().size());
      Vector2D newPoint = game.getDescription().getTransforms().get(j).transform(game.getCurrentPoint());
      game.setCurrentPoint(game.findValidPoint(newPoint));
      game.getCanvas().putPixel(game.getCurrentPoint(), false);
    }
    return null;
  }

  // Step 2, 3, 4: Divide the task into smaller tasks and execute them in parallel
  public void runSteps(int n) {
    int numProcessors = Runtime.getRuntime().availableProcessors();
    ExecutorService executor = Executors.newFixedThreadPool(numProcessors);

    int chunkSize = n / numProcessors;
    List<Future<Void>> futures = new ArrayList<>();

    for (int i = 0; i < numProcessors; i++) {
      int start = i * chunkSize;
      int end = (i == numProcessors - 1) ? n : start + chunkSize; // handle remainder
      Callable<Void> task = new ChaosGameTask(this.game, start, end);
      futures.add(executor.submit(task));
    }

    for (Future<Void> future : futures) {
      try {
        future.get(); // wait for task to complete
      } catch (InterruptedException | ExecutionException e) {
        e.printStackTrace();
      }
    }

    executor.shutdown(); // always remember to shutdown the executor
  }
}