package edu.ntnu.stud.chaosgame.view;
import javafx.scene.image.PixelWriter;
import javafx.scene.image.WritableImage;
import javafx.scene.paint.Color;
import edu.ntnu.stud.chaosgame.controller.game.ChaosCanvas;

/**
 * This class converts the state of a ChaosCanvas to a WritableImage.
 */
public class ChaosCanvasToImageConverter {

  /**
   * The image to be created.
   */
  private WritableImage image;

  /**
   * Convert the canvas to a writable image.
   *
   * @param chaosCanvas the canvas to work upon.
   */
  public ChaosCanvasToImageConverter(ChaosCanvas chaosCanvas, boolean useIntensity) {
    if (!useIntensity) {
      this.convertWithoutIntensity(chaosCanvas);
    } else {
      this.convertWithIntensity(chaosCanvas);
    }
  }


  /**
   * Get the image.
   *
   * @return the image.
   */
  public WritableImage getImage() {
    return image;
  }

  /**
   * Convert the ChaosCanvas to an image using the intensity of the pixels
   * to determine the color.
   *
   * @param canvas the ChaosCanvas to convert.
   */
  public void convertWithIntensity(ChaosCanvas canvas) {
    int width = canvas.getWidth();
    int height = canvas.getHeight();
    image = new WritableImage(width, height);
    PixelWriter pixelWriter = image.getPixelWriter();

    int[][] canvasArray = canvas.getCanvasArray();
    int[][] canvasIntensityArray = canvas.getCanvasIntensityArray();
    int maxIntensity = 0;

    // Find the maximum intensity.
    for (int i = 0; i < width; i++) {
      for (int j = 0; j < height; j++) {
        if (canvasArray[i][j] == 1 && canvasIntensityArray[i][j] > maxIntensity) {
          maxIntensity = canvasIntensityArray[i][j];
        }
      }
    }

    // Map the intensity to a hue value.
    for (int i = 0; i < width; i++) {
      for (int j = 0; j < height; j++) {
        if (canvasArray[i][j] == 1) {
          int intensity = canvasIntensityArray[i][j];
          double hue = (double) intensity / maxIntensity * 360; // Map intensity to hue
          pixelWriter.setColor(j, i, Color.hsb(hue, 1.0, 1.0)); // Full saturation and brightness
        } else {
          pixelWriter.setColor(j, i, Color.WHITE);
        }
      }
    }
  }

  /**
   * Convert the ChaosCanvas to an image without concern for the intensity of the pixels.
   *
   * @param chaosCanvas the ChaosCanvas to convert.
   */
  public void convertWithoutIntensity(ChaosCanvas chaosCanvas) {
    int width = chaosCanvas.getWidth();
    int height = chaosCanvas.getHeight();
    image = new WritableImage(width, height);
    PixelWriter pixelWriter = image.getPixelWriter();

    int[][] canvasArray = chaosCanvas.getCanvasArray();
    for (int i = 0; i < width; i++) {
      for (int j = 0; j < height; j++) {
        if (canvasArray[i][j] == 1) {
          pixelWriter.setColor(j, i, Color.BLACK);
        } else {
          pixelWriter.setColor(j, i, Color.WHITE);
        }
      }
    }
  }

}
